/* --------------------------------------------------------------------------
 * Copyright 1992-1994 by Forschungszentrum Informatik (FZI)
 * All rights reserved.
 *
 * You can use and distribute this software under the terms of the license
 * you should have received along with this software; either version 1.1 of
 * the license, or (at your option) any later version.
 * For a copy of the license or for additional information about this software,
 * write to Xcc Software, Durlacher Allee 53, D-76131 Karlsruhe, Germany;
 * Email: obst@xcc-ka.de.
 * --------------------------------------------------------------------------
 */
/* OBST LIBRARY MODULE */
/* ========================================================================= */
/* MODULE IMPLEMENTATION                                                     */
/* ========================================================================= */
/*                                                                           */
/* MODULE : extended syncronization                                          */
/*          all that goes beyond locking (currently optimistic and timestamp)*/
/*                                                                           */
/* ORIGINAL                                                                  */
/* AUTHOR: Axel Freyberg, Michael Ranft        DATE: 29.09.1993              */
/*                                                                           */
/* CHANGES: see .changes                                                     */
/*                                                                           */
/* VERSION: none                                                             */
/* ========================================================================= */

#define OBST_IMP_MEMOP
#include "_obst_config.h"
#include "obst_stdinc.h"
 
#include "obst.h"
#include "obst_progstd.h"

#include "sync_obst.h"
#include "sync_decl.h"
#include "sync_err.h"
#include "sync_trc.h"
#include "ext_sync_obst.h"

#include "trc_tsy.h"

#include "psm_util.h"


/* ========================================================================= */
/* CLASS IMPLEMENTATION                                                      */
/* ========================================================================= */
/*                                                                           */
/* CLASS-NAME: ExtSyncObj                                                    */
/*                                                                           */
/* ORIGINAL                                                                  */
/* AUTHOR: Michael Ranft                       DATE: 29.09.1993              */
/*                                                                           */
/* ========================================================================= */


/* ================================================================ */
/* Method: ExtSyncObj::get                                          */
/*                                                                  */
/* Error: SYNC_OK, SYNC_WRONG_ACCESS, SYNC_TA_ABORTED,              */
/*        SYNC_IMPLTA_ABORTED                                       */
/* ================================================================ */

sos_Bool ExtSyncObj::get (sos_Access_mode am, sos_Sync_mode sm)

{ 
  T_PROC("ExtSyncObj::get");
  TT(tsy_H, T_ENTER);
  TT(tsy_L, TAM(am); TSM(sm));
  TSOBJ(self);

  SyncCont         syct;
  Transaction      ta; 
  LockingTA        tplta;
  sos_HostProc_Id  hpid;
  sos_HostProc_Id  tapid;
  sos_Open_result  result;
  sos_Bool         access_delay = TRUE;
  sos_Bool         okay;
  sos_Bool         impl_firstcont = FALSE;
  ta_Access_mode   ct_am;

  // control parameter 
  if ((am != READING) && (am != WRITING))
    { self.set_error(SYNC_WRONG_ACCESS);
      TT(tsy_H, TXT("SYNC_WRONG_ACCESS"); T_LEAVE);
      return FALSE; }

  while (access_delay)
    {     
      access_delay = FALSE; 
       
      syct = SyncCont::get_root();

      // look for transaction 
      hpid = syct.get_HP();
      ta  = syct.get_ta_tab()[hpid];
      TT(tsy_VL, TI(hpid));
      TTA(ta);

      // is transaction intern aborted ?
      if (VALID(ta) && (ta.get_status() == ABORTED))
        { // nothing to do
          self.set_error(SYNC_OK);
          TT(tsy_H, TXT("nothing to do -> TRUE"); T_LEAVE);
          return TRUE; }
 

      if ((self.get_awaking()) &&
          (INVALID(ta) || (ta.get_type() != OPTIMISTIC)))
        { // another ta was waiting to access this SyncObj and to this ta a
          // message was sent that it can access the SyncObj. But as long
          // as awaking is set, the ta has not accessed the SyncObj, it has not
          // awake. The other ta, which has accessed the SyncObj before has 
          // probably released all locks, so that to keep evrything in order
          // we have to wait, till the other ta has accessed the Obj, 
          // under the assumption that the process was not aborted.
          
          if (!self.get_queue().is_empty())
            {    
              tapid = self.get_queue().get_nth(1).get_ta().get_hostproc_id();
              okay  = syct.process_check(tapid);
              if (!okay) 
                 syct.process_remove(tapid);
              else
              { // close Sync_Container and open it to give the awaking ta
                // the possibility to access the Sync_Container
                
                SYNC_CONTAINER.close();
      
                result=SYNC_CONTAINER.open(WRITING, WAITING);
                if (result != OPENED)
                   err_raise(err_SYS, err_SYNC_NO_SYNC_CONTAINER, ERR, FALSE);

                access_delay = TRUE;
              } 
            }
            else
              self.set_awaking(FALSE); 
        } 
    }  

  // if no transaction, create one implicit
  if INVALID(ta) 
    { TT(tsy_L, TXT("create implicit TA"));
      ta = ExtLockingTA::create(SYNC_CONTAINER);
      if INVALID(ta) 
        { TT(tsy_H, TXT("SYNC_NO_CREATE_IMPLICIT_TA -> FALSE"); T_LEAVE);
          err_raise(err_SYS, err_SYNC_NO_CREATE_IMPLICIT_TA, ERR, FALSE); 
        }
      ta.set_implicit(TRUE);
      impl_firstcont = TRUE;
      TTA(ta);
    }
 

  // container already in set of transaction ?
  ct_am = ta.is_accessed (self);
  TT(tsy_VL, TXT("CT:"); TTM(ct_am); TXT("AM:"); TAM(am));

  // change to higher access
  if ((ct_am == READ_AC) && (am == WRITING))
    {
      if (ta.get_type() == LOCKING)
        { okay = self.tpl_writelock(LockingTA::make(ta), sm);
          if (!okay)
            { TT(tsy_H, TXT("FALSE"); T_LEAVE); 
              return FALSE; } 
        }     

      // move from readset to writeset
      ta.get_SO (self, am);
      self.set_error(ta.get_error());
      TT(tsy_H, TB(self.get_error()==SYNC_OK); T_LEAVE);
      return sos_Bool(self.get_error() == SYNC_OK);
    }

      
  // first access to container by transaction
  if ((ct_am == NO_AC) && ((am == READING) || (am == WRITING)))
    { // call different methods for syncronization
      switch (ta.get_type()) 
        { 
          case LOCKING:
              { if (am == READING) 
                  okay = self.tpl_readlock(LockingTA::make(ta), sm);
                 else
                  okay = self.tpl_writelock(LockingTA::make(ta), sm); 
                break; }
          case TIMESTAMP:
              { okay = self.tso_read(TimestampTA::make(ta), sm);
                break; }
          case OPTIMISTIC:
              { if (am == READING)
                  self.opt_read(OptimisticTA::make(ta));
                 else
                  self.opt_write(OptimisticTA::make(ta));
                okay = TRUE; 
                break; }
        }


      // if read failed, return
      if (!okay) 
        { // implict transaction objects have to be destroyed if the 
          // transaction was intern aborted or if the implict transaction just
          // has been started at the beginning of this method and the SyncObj,
          // which should be opened TESTING, failed to open.
          if ((ta.get_implicit()) && 
             ((impl_firstcont) || (self.get_error() == SYNC_IMPLTA_ABORTED)))
            { TT(tsy_L, TXT("Implicit TA intern aborted -> destroy"));
              tplta = LockingTA::make(ta);
              tplta.destroy(); }
          TT(tsy_H, TXT("FALSE"); T_LEAVE);
          return FALSE; } 

      ta.get_SO (self, am);
      self.set_error(ta.get_error());
      TT(tsy_H, TB(self.get_error() == SYNC_OK); T_LEAVE);
      return sos_Bool(self.get_error() == SYNC_OK);       
    }      



  // the other cases: ct_am = READ_AC && am == READING
  //                  ct_am = WRITE_AC && (am == READING || am == WRITING)

  // reopen if already released
  ta.get_SO(self, am);
  if (ta.get_error() != SYNC_OK)
    { self.set_error(ta.get_error());
      TT(tsy_H, TXT("FALSE"); T_LEAVE);
      return FALSE; }   


  // access is already given
  self.set_error(SYNC_OK);
  TT(tsy_H, T_LEAVE);
  return TRUE;      
}   

/* ================================================================ */
/* Method: ExtSyncObj::release()                                    */
/*                                                                  */
/* Error: SYNC_OK, SYNC_NO_GET, SYNC_TA_ABORTED (ts),               */
/* ================================================================ */

void ExtSyncObj::release ()

{ 
  T_PROC("ExtSyncObj::release");
  TT(tsy_H, T_ENTER);
  TSOBJ(self);

  SyncCont        syct;
  Transaction     ta;
  LockingTA       tplta; 
  sos_HostProc_Id hpid;
  sos_HostProc_Id tapid;
  sos_Open_result result;
  sos_Bool        access_delay = TRUE;
  sos_Bool        okay;
  ta_Access_mode  ct_am;


  while (access_delay)
    { 
      access_delay = FALSE;

      syct = SyncCont::get_root();

      // look for transaction 
      hpid = syct.get_HP();
      ta  = syct.get_ta_tab()[hpid];  
      TT(tsy_VL, TI(hpid));
      TTA(ta);

      // if no transaction, no container open
      if INVALID(ta) 
        { self.set_error(SYNC_NOT_GET);
          TT(tsy_H, TXT("SYNC_NOT_GET"); T_LEAVE);
          return; }


      // is transaction intern aborted ?
      if (ta.get_status() == ABORTED)
        { // nothing to do
          self.set_error(SYNC_OK);
          TT(tsy_H, TXT("intern aborted - nothing to do"); T_LEAVE);
          return; }


      if ((self.get_awaking()) &&
          (ta.get_type() != OPTIMISTIC))
        { // another ta was waiting to access this SyncObj and to this ta a
          // message was sent that it can access the SyncObj. But as long
          // as awaking is set, the ta has not accessed the SyncObj, it has not
         // awake. The other ta, which has accessed the SyncObj before has
          // probably released all locks, so that to keep evrything in order
          // we have to wait, till the other ta has accessed the Obj,
          // under the assumption that the process was not aborted.
         
          if (!self.get_queue().is_empty())
            {
              tapid = self.get_queue().get_nth(1).get_ta().get_hostproc_id();
              okay  = syct.process_check(tapid);
              if (!okay)
                 syct.process_remove(tapid);
              else
              { // close Sync_Container and open it to give the awaking ta
                // the possibility to access the Sync_Container
 
                SYNC_CONTAINER.close();
 
                result=SYNC_CONTAINER.open(WRITING, WAITING);
                if (result != OPENED)
                   err_raise(err_SYS, err_SYNC_NO_SYNC_CONTAINER, ERR, FALSE);
 
                access_delay = TRUE;
              }
            }
            else
              self.set_awaking(FALSE);
        }
    } 

  // get access mode
  ct_am = ta.is_accessed(self);
  TT(tsy_L, TXT("CT:"); TTM(ct_am));
  if (ct_am == NO_AC)
    { self.set_error(SYNC_NOT_GET);
      TT(tsy_H, TXT("SYNC_NOT_GET"); T_LEAVE);
      return; }


  // call different methods for syncronisation
  if ((ta.get_type() == TIMESTAMP) && (ct_am == WRITE_AC) &&
      (!ta.is_released(self)))
    { // write for Timestamp
      self.tso_write(TimestampTA::make(ta));
      if (self.get_error() != SYNC_OK)
        { TT(tsy_H, T_LEAVE);
          return; }
    } 

  // remove from openset
  ta.release_SO(self);
  if (ta.get_error() != SYNC_OK)
    { self.set_error(ta.get_error());
      TT(tsy_H, T_LEAVE);
      return; }


  // commit implicit transactions
  // !! beware of  _TA aborted implicitly somewhere before 
  TT(tsy_VL, TI(ta.get_openset().card()));
  if (ta.get_implicit())
    { if (ta.get_openset().is_empty())
	{ tplta = LockingTA::make(ta);
	  tplta.commit();
	  tplta.destroy();
	}
      else
      if (!ta.get_writeset().is_empty())
	{ sos_SOSet helpset = sos_SOSet::create(TEMP_CONTAINER);
          helpset += ta.get_openset();
          helpset *= ta.get_writeset();
          if (helpset.is_empty())
            { tplta = LockingTA::make(ta);
              tplta.commit();
            } 
        }
    }
  

  // end release
  self.set_error(SYNC_OK);
  TT(tsy_H, T_LEAVE);
  return;
}  

